/*
 * Most Recent Version 29-oct-97:
 *  - no more debugging options etc
 *  - choice of 5 different pictures (07-nov-97)
 *
 *
 * Fractal  UNINSTRUMENTIERT
 *
 * Parallel Recursive Benchmark - compute Mandelbrot set using
 * "if edges have same color, fill it" heuristic and Divide&Conquer
 * (computes e.g. just 95886 pixels on a 512x512 fractal instead of 262144 in
 * the brute force approach)
 *
 *
 * Usage:
 *
 * fractal  [x_pixel [y_pixel [recurs]]]
 *
 *		x_pixel	= number of pixels to compute (x)
 *		y_pixel = dito for y
 *		recurs	= 0 for naive algorithm, otherwise recursive D&C
 *
 *
 * Stefan U. Haenssgen  27-jan-97
 *
 *
 * 27-jan-97	1st version, rough implementation
 *		Use x1,y1-x2,y2 instead of x,y,size
 *		Use independent x/y size
 *		Use recursive/naive algorithm dependent on argv
 * 28-jan-97	Debug level flexible, output ASCII data if large enough etc
 * 29-jan-97	Added instrumentation and profiling
 *		 (ProfileAddStats, ...)
 *		Added Thread creation, predicates etc
 * 31-jan-97	Added timing around computation itself, i.e. don't
 *		 measure malloc!
 *		Removed pixel counter, isn't mutex protected, may cause
 *		 write conflicts, just ignore it...
 *
 */



#include <stdio.h>

#include <sys/time.h>
#include <sys/timeb.h>



#define MAXITER 1024			/* Maximum recursion depth	*/

unsigned char *pixel;			/* Pixel values 		*/

/* Upper left coordinates and Coordinate size (x and y) */

#define COORDS 1

#if COORDS == 1
double xcorner = -2.25;                 /* basic fractal, symmetrical   */
double ycorner = -1.5;
double csize   =  3.0;
#endif
#if COORDS == 2
double xcorner = -0.058500;             /* much work in lower left corner*/
double ycorner = -0.885000;
double csize   =  0.120000;
#endif
#if COORDS == 3
double xcorner = -0.013095;             /* only lower right corner      */
double ycorner = -0.810949;
double csize   =  0.000038;
#endif
#if COORDS == 4
double xcorner = -0.538020;             /* non-symmetrical but equal work*/
double ycorner = -0.528120;
double csize   =  0.004800;
#endif
#if COORDS == 5
double xcorner = -0.764412;             /* little work in lower right   */
double ycorner = -0.101004;
double csize   =  0.000960;
#endif

int    pxsize  = 60;			/* Pixel number x, y		*/
int    pysize  = 30;
double gap     =  0.1;			/* Distance between screen pixels */
int    iter    = MAXITER;		/* Iteration limit		*/

int    use_recursion = 1;		/* Use recursive algorithm?	*/

/*long cp = 0;				/* Computed pixels counter	*/



/* Declare recursive function here */

void compute_area(int x1, int y1, int x2, int y2);


/* Output graphics as ASCII picture on stdout */

void print_pixels()
{
    int x,y;
	
    for (y=0; y<pysize; y++) {
	for (x=0; x<pxsize; x++) {
	    putchar(pixel[x+pxsize*y]%64+32);
	}	
	putchar('\n');
	fflush(stdout);
    }
}
    

/* Compute one pixel at the given pixel coordinates */

unsigned char compute_pixel(int x, int y)
{
    int count;
    double xc, yc, xt, yt, xz, yz, t;

    /*cp++;*/
    
    xc = xcorner + ((double)x)*gap;
    yc = ycorner + ((double)y)*gap;
    
    xt=yt=xz=yz=count=0;
    while ((++count<iter) && (xt+yt<4)) {
	t  = xz*yz;
	yz = t + t + yc;
	xz = xt-yt+xc;
	xt = xz*xz;
	yt = yz*yz;
    }

    return((unsigned char) count % 256);
}


/* Compute edges of area (x1,y1) to (x2, y2) */
/* size has to be >= 1 */

void compute_frame(int x1, int y1, int x2, int y2)
{
    int xx,yy;

    yy=y1;
    for (xx=x1; xx<=x2; xx++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    yy=y2;
    for (xx=x1; xx<=x2; xx++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    xx=x1;
    for (yy=y1+1; yy<=y2-1; yy++)	/* (don't compute corners twice) */
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    xx=x2;
    for (yy=y1+1; yy<=y2-1; yy++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    
}


/* Compute cross through area */
/* size has to be >= 1, don't re-compute edge pixels */

void compute_cross(int x1, int y1, int x2, int y2)
{
    int xx,yy,y0;

    yy=(y1+y2)/2;
    for (xx=x1+1; xx<=x2-1; xx++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    y0=yy;			/* don't compute middle pixel twice */
    xx=(x1+x2)/2;
    for (yy=y1+1; yy<=y0-1; yy++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    for (yy=y0+1; yy<=y2-1; yy++)
	pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
    
}


/* True if edges of area are same color */

char edges_match(int x1, int y1, int x2, int y2)
{
    int 	  xx,yy;
    char          match = 1;
    unsigned char color = pixel[x1+pxsize*y1];	/* upper left corner 	*/

    xx=x1;
    yy=y1;
    while ((match) && (xx<=x2)) {
	match = (color == pixel[xx+pxsize*yy]);
	xx++;
    }
    xx=x1;
    yy=y2;
    while ((match) && (xx<=x2)) {
	match = (color == pixel[xx+pxsize*yy]);
	xx++;
    }
    xx=x1;
    yy=y1+1;
    while ((match) && (yy<=y2-1)) {
	match = (color == pixel[xx+pxsize*yy]);
	yy++;
    }
    xx=x2;
    yy=y1+1;
    while ((match) && (yy<=y2-1)) {
	match = (color == pixel[xx+pxsize*yy]);
	yy++;
    }

    return(match);
}


/* Compute area (x1,y1) to (x2,y2) */
/* Fill area if all edges have same color */
/* Don't touch already computed areas (i.e. edges) */

void compute_area(int x1, int y1, int x2, int y2)
{
    unsigned char p;
    int           xx,yy;
    int           dx = x2-x1;
    int           dy = y2-y1;

    
    if ((dx < 2) || (dy < 2)) {
	return;
    }
    if (dx == 2) {			/* Handle single-line cases	*/
	xx=(x1+x2)/2;
	for (yy = y1+1; yy <= y2-1; yy++)
	    pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
	return;
    }
    if (dy == 2) {
	yy=(y1+y2)/2;
	for (xx = x1+1; xx <= x2-1; xx++)
	    pixel[xx+pxsize*yy] = compute_pixel(xx,yy);
	return;
    }
    if (edges_match(x1,y1,x2,y2)) {	/* If edges match, fill area - done! */
	
	p = pixel[x1+pxsize*y1];
	
	for (yy=y1+1; yy<=y2-1; yy++) {
	    for (xx=x1+1; xx<=x2-1; xx++) {
		pixel[xx+pxsize*yy] = p;
	    }
	}
	return;
    }

    /* Subdivide area into quadrants, compute them recursively 		*/
    
    compute_cross(x1,y1,x2,y2);
    
    compute_area(x1, y1, (x1+x2)/2, (y1+y2)/2);
    compute_area((x1+x2)/2, y1, x2, (y1+y2)/2);
    compute_area(x1, (y1+y2)/2, (x1+x2)/2, y2);
    compute_area((x1+x2)/2, (y1+y2)/2, x2, y2); /* NOTHREAD */
    
}


main(int argc, char **argv)
{
    struct timeval starttime;
    struct timeval endtime;
    int x,y;
    int i;

    
    i=1;
    if (argc>i) {
	pxsize = atoi(argv[i]);
    }
    i++;
    if (argc>i) {
	pysize = atoi(argv[i]);
    }
    i++;
    if (argc>i) {
	use_recursion = atoi(argv[i]);
    }
    i++;

    printf("Fractal: %d x %d\n", pxsize, pysize);

    gap = csize/pxsize;			/* Distance between screen pixels */

    /* Allocate & initialize pixel buffer */
    
    pixel = (unsigned char*) malloc(sizeof(unsigned char) *
				    pxsize * pysize + 1);
    for (x=0; x<pxsize*pysize; x++)
	pixel[x]=(unsigned char)0;


    gettimeofday(&starttime,NULL);
    
    if (use_recursion) {

	printf("Using recursive algorithm\n");
	
	compute_frame(0,0,pxsize-1,pysize-1);

	compute_area(0,0,pxsize-1,pysize-1);

    } else {

	printf("Using naive (non-recursive) algorithm\n");
    
	for (y=0; y<pysize; y++) {
	    for (x=0; x<pxsize; x++) {
		pixel[x+pxsize*y] = compute_pixel(x,y);
	    }
	}
    
    }

    gettimeofday(&endtime,NULL);

    printf("Elapsed time %f\n",
	   (endtime.tv_sec - starttime.tv_sec) +
	   (endtime.tv_usec - starttime.tv_usec)/1000000.0 );

    /*printf("%d pixels computed\n",cp);*/
    
    exit(0);

}
