/* -*- C -*- */
/*
 * Heat diffusion (Jacobi-type iteration)
 *
 * Usage: 1) heat reads a parameter file (heatparams - hard coded)
 *        2) Specify the granularity of the Cilk threads by setting
 *           leafmaxcol via a command line argument, e.g.
 *
 *              heat -nproc 7 16
 *
 *           Here, the maximum number of columns of a Cilk thread 
 *           will be "16". Given nx = 1024 columns, there will be
 *           1024/16 = 64 threads executing the five-point-stencil
 *           computation. 
 *
 *
 *
 * Volker Strumpen, Boston                                 August 1996
 *
 * $Id: heat.cilk,v 1.2 1997/02/15 00:11:26 strumpen Exp $
 *
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *
 *
 *
 * REAPAR Version by Stefan U. Haenssgen
 *
 *
 * 28-jan-98	De-Cilk-ified, removed keywords and special functions
 *		Minor reformatting
 *		Removed sync/spawn
 *		Made all recursive procedures return void, use reference
 *		 parameters instead
 *		Added needresults annotation
 *		Take parameters from stdin, not from fixed file
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <math.h>
#include <errno.h>
extern int errno;


/* Define ERROR_SUMMARY if you want to check your numerical results */
#undef ERROR_SUMMARY


#define f(x,y)     (sin(x)*sin(y))
#define randa(x,t) (0.0)
#define randb(x,t) (exp(-2*(t))*sin(x))
#define randc(y,t) (0.0)
#define randd(y,t) (exp(-2*(t))*sin(y))
#define solu(x,y,t) (exp(-2*(t))*sin(x)*sin(y))

void read_heatparams(void);

int nx, ny, nt;
double xu, xo, yu, yo, tu, to;
double dx, dy, dt;

double dtdxsq, dtdysq;
double t;

int leafmaxcol;

/*****************   Allocation of grid partition  ********************/

void allcgrid(double **new, double **old, int lb, int ub)
{
  int j;
  double **rne, **rol;

  for (j=lb, rol=old+lb, rne=new+lb; j < ub; j++, rol++, rne++) {
    *rol = (double *) malloc(ny * sizeof(double));
    *rne = (double *) malloc(ny * sizeof(double));
  }
}

/*****************   Initialization of grid partition  ********************/

void initgrid(double **old, int lb, int ub)
{
  int a, b, llb, lub;

  llb = (lb == 0) ? 1 : lb;
  lub = (ub == nx) ? nx - 1 : ub;
  
  for (a=llb, b=0; a < lub; a++)		/* boundary nodes */
    old[a][b] = randa(xu + a * dx, 0);
  
  for (a=llb, b=ny-1; a < lub; a++)
    old[a][b] = randb(xu + a * dx, 0);
  
  if (lb == 0) {
    for (a=0, b=0; b < ny; b++)
      old[a][b] = randc(yu + b * dy, 0);
  }
  if (ub == nx) {
    for (a=nx-1, b=0; b < ny; b++)
      old[a][b] = randd(yu + b * dy, 0);
  }
  for (a=llb; a < lub; a++) {	/* inner nodes */
    for (b=1; b < ny-1; b++) {
      old[a][b] = f(xu + a * dx, yu + b * dy);
    }
  }
}


/***************** Five-Point-Stencil Computation ********************/

void compstripe(register double **new, register double **old, int lb, int ub)
{
  register int a, b, llb, lub;
  
  llb = (lb == 0) ? 1 : lb;
  lub = (ub == nx) ? nx - 1 : ub;
  
  for (a=llb; a < lub; a++) {
    for (b=1; b < ny-1; b++) {
      new[a][b] =   dtdxsq * (old[a+1][b] - 2 * old[a][b] + old[a-1][b])
	          + dtdysq * (old[a][b+1] - 2 * old[a][b] + old[a][b-1])
  	          + old[a][b];
    }
  }
  
  for (a=llb, b=ny-1; a < lub; a++)
    new[a][b] = randb(xu + a * dx, t);
  
  for (a=llb, b=0; a < lub; a++)
    new[a][b] = randa(xu + a * dx, t);
  
  if (lb == 0) {
    for (a=0, b=0; b < ny; b++)
      new[a][b] = randc(yu + b * dy, t);
  }
  if (ub == nx) {
    for (a=nx-1, b=0; b < ny; b++)
      new[a][b] = randd(yu + b * dy, t);
  }
}


/***************** Decomposition of 2D grids in stripes ********************/

#define ALLC       0
#define INIT       1
#define COMP       2

/*cilk*/
void divide(int *retval, int lb, int ub,
	    double **new, double **old, int mode, int timestep)
{
  int l, r;

  if (ub - lb > leafmaxcol) {
    divide(&l, lb, (ub + lb) / 2, new, old, mode, timestep);
    divide(&r, (ub + lb) / 2, ub, new, old, mode, timestep); /* NOTHREAD */
    /* NEEDRESULTS */
    *retval = l + r;
    return;
  } else {
    switch (mode) {
    case COMP: {
      if (timestep % 2)
	compstripe(new, old, lb, ub);
      else
	compstripe(old, new, lb, ub);
      *retval = 1;
      return;
    }
      
    case ALLC: {
      allcgrid(new, old, lb, ub);
      *retval = 1;
      return;
    }
      
    case INIT: {
      initgrid(old, lb, ub);
      *retval = 1;
      return;
    }
    }
  }
  *retval = 0;
  return;
}


/*cilk*/
int heat(void)
{
  double **old, **new, **mat;
  double tmp;
  double mae = 0.0;
  double mre = 0.0;
  double me = 0.0;
  int a, b, c, l;
  
  
  /* Memory Allocation */
  old = (double **) malloc(nx * sizeof(double *));
  new = (double **) malloc(nx * sizeof(double *));
  divide(&l, 0, nx, new, old, ALLC, 0);

  /* 
   * Sequential allocation might be faster than parallel!
   * Substitute the following call for the preceding divide:
   *
   * allcgrid(new, old, 0, nx);
   */
  printf("\n Allocation complete.");
     
  /* Initialization */
  divide(&l, 0, nx, new, old, INIT, 0);

  printf("\n\n Number of leaf nodes:            %10d\n", l);
  
  
  /* Jacobi Iteration (divide x-dimension of 2D grid into stripes) */

  for (c = 1; c <= nt; c++) {
    t = tu + c * dt;
    
    divide(&l, 0, nx, new, old, COMP, c);

  }


#ifdef ERROR_SUMMARY  
  /* Error summary computation: Not parallelized! */
  mat = (c % 2) ? old : new;
  printf("\n Error summary of last time frame comparing with exact solution:");
  for (a=0; a<nx; a++)
    for (b=0; b<ny; b++) {
      tmp = fabs(mat[a][b] - solu(xu + a * dx, yu + b * dy, to));
      if (tmp > mae)
	mae = tmp;
    }
  printf("\n   Local maximal absolute error  %10e ", mae);
  
  for (a=0; a<nx; a++)
    for (b=0; b<ny; b++) {
      tmp = fabs(mat[a][b] - solu(xu + a * dx, yu + b * dy, to));
      if (mat[a][b] != 0.0)
	tmp = tmp / mat[a][b];
      if (tmp > mre)
	mre = tmp;
    }
  printf("\n   Local maximal relative error  %10e %s ", mre * 100, "%");
  
  me = 0.0;
  for (a=0; a<nx; a++)
    for (b=0; b<ny; b++) {
      me += fabs(mat[a][b] - solu(xu + a * dx, yu + b * dy, to));
    }
  me = me / (nx * ny);
  printf("\n   Global Mean absolute error    %10e\n\n", me);
#endif
  return 0;
}


int main(int argc, char *argv[])
{
  int ret;

  if (argc != 2) {
    printf("\nUsage: %s [Cilk options] <minimum number of columns in leaf node partition>\n\n",
	   argv[0]);
    exit(-1);
  }
  leafmaxcol = atoi(argv[1]);
  
  read_heatparams();
  
  ret = heat();
  
  return 0;
}

void read_heatparams(void)
{
  char filefn[128];
  FILE *f;
  int l;
  
  printf("\n Reading input parameters ...\n");
  /* 
   *  Filename is just "heatparams". Change as desired!
   */
/*  sprintf(filefn, "%s", "heatparams");*/
/*  if ((f = fopen(filefn, "r")) == NULL) {
    printf("\n Can't open %s\n", filefn);
    exit(0);
  }
  */
  l = scanf("%d %d %d %lf %lf %lf %lf %lf %lf",
	     &nx, &ny, &nt, &xu, &xo, &yu, &yo, &tu, &to);
  if (l != 9)
    printf("\n Warning: fscanf errno %d", errno);
/*  fclose(f);*/
  
  printf("\n Application Data Block is:");
  printf("\n   nx   %d", nx);
  printf("\n   ny   %d", ny);
  printf("\n   nt   %d", nt);
  printf("\n   xu   %f", xu);
  printf("\n   xo   %f", xo);
  printf("\n   yu   %f", yu);
  printf("\n   yo   %f", yo);
  printf("\n   tu   %f", tu);
  printf("\n   to   %f", to);
  
  dx = (xo - xu) / (nx - 1);
  dy = (yo - yu) / (ny - 1);
  dt = (to - tu) / nt;	/* nt effective time steps! */
  
  printf("\n   dx = %f", dx);
  printf("\n   dy = %f", dy);
  printf("\n   dt = %f", dt);
  
  printf("\n\n Stability Value for explicit method must be > 0:  %f\n",
	 0.5 - (dt / (dx * dx) + dt / (dy * dy)));
  
  dtdxsq = dt / (dx * dx);
  dtdysq = dt / (dy * dy);
}
