/* MERGED all programs as well as the header files... 	*/
/* NOPARALLEL annotations for build procedures		*/

/* For copyright information, see olden_v1.0/COPYRIGHT */

/*
 * power.h
 *
 * By:  Martin C. Carlisle
 * 6/15/94
 *
 * Header file for the Power Pricing problem
 * adapted from code by Steve Lumetta, Sherry Li, and Ismail Khalil
 * University of California at Berkeley
 *
 *
 * Adapted by Stefan U. Haenssgen 15-dec-96
 *
 * 15-dec-96	Random degree varying from n/2..3n/2 for original fixed
 *		 degree n -> my_num_XXX in all structs and maximum
 *		 size of Arrays *2
 *		Thread stuff added
 * 22-dec-96	Increased remaining array (leafs) to 3*size
 * 29-dec-96	Loss factors etc as #defined parameters
 *		Increased array to 5*size
 * 14-jul-97	Added all code together into one giant file
 * 16-jul-97	Rewrote result passing from function value
 *		 to reference parameter passing
 * 18-jul-97	Added parallelization annotations
 * 05-sep-97	Added ";" before #directives to make skip-to-
 *		 beginning-of-statement safer
 * 
 */

#include <errno.h>


typedef struct demand {
  double P;
  double Q;
} Demand;


#include "math.h"
extern void *malloc();

/*#define NULL 0*/

#define NUM_FEEDERS 10
#define LATERALS_PER_FEEDER 20
#define BRANCHES_PER_LATERAL 5
#define LEAVES_PER_BRANCH 10


#define F_EPSILON 0.000001
#define G_EPSILON 0.000001
#define H_EPSILON 0.000001
#define ROOT_EPSILON 0.00001

/* Root price multipliers */

#define TREE_THETA_R 0.7		/* 0.8 */
#define TREE_THETA_I 0.14		/* 0.16 */

/* Lines to lateral nodes */

#define LATERAL_X 0.000001		/* 0.000001 */
#define LATERAL_R  1/3000000.0 		/* 1/300000.0 */
#define LATERAL_ALPHA 0.0
#define LATERAL_BETA 0.0

/* Lines to branches */

#define BRANCH_R 0.00001 		/* 0.0001 */
#define BRANCH_X 0.00002 		/* 0.00002 */
#define BRANCH_ALPHA 0.0
#define BRANCH_BETA 0.0


typedef struct root {
  Demand D;
  double theta_R; 
  double theta_I; 
  Demand last;
  double last_theta_R; 
  double last_theta_I;
  struct feeder *feeder;
} *Root;

typedef struct feeder {
  Demand D;
  struct feeder *next_feeder;
  struct lateral *lateral;
} *Feeder;

typedef struct lateral {
  Demand D;
  double alpha;
  double beta;
  double R;
  double X;
  struct lateral *next_lateral;
  struct branch *branch;
} *Lateral;

typedef struct branch {
  Demand D;
  double alpha;
  double beta;
  double R;
  double X;
  struct branch *next_branch;
  int    my_leaves_per_branch;
  struct leaf *leaves[5*LEAVES_PER_BRANCH];
} *Branch;

typedef struct leaf {
  Demand D;
  double pi_R;
  double pi_I;
} *Leaf;

#define chatting printf
#define __MyNodeId 0

#include <fcntl.h>

int flag;
int __NumNodes;
int __NDim;

int num_feeders;
int laterals_per_feeder;
int branches_per_lateral;
int leaves_per_branch;


/* Prototypes */
Root build_tree();
Feeder build_feeder(int i, int num);
Lateral build_lateral(int i, int num);
Branch build_branch(int i, int j, int num);
Leaf build_leaf();

void Compute_Tree(Root r);
/*Demand Compute_Lateral(Lateral l, double theta_R, double theta_I,
                       double pi_R, double pi_I, int depth);
Demand Compute_Branch(Branch b, double theta_R, double theta_I,
                       double pi_R, double pi_I, int depth);
Demand Compute_Leaf(Leaf l, double pi_R, double pi_I);*/


#define max(x,y) (x>y?x:y)



/* compute.c		WITHOUT PROFILING / THREADS
 *
 * By:  Martin C. Carlisle
 * 6/15/94
 *
 * Implements computation phase of the Power Pricing problem
 * based on code by: Steve Lumetta, Sherry Li, and Ismail Khalil
 * University of California at Berkeley
 *
 *
 * Adapted by Stefan U. Haenssgen  29-nov-96
 *
 * 29-nov-96	Removed futures, CM-specifics, etc.
 *		Problem sizes as command line parameters
 *		Instrumentation for profiling (just recursion)
 * 15-dec-96	Random degree / number of leaves etc, varying
 *		 between n/2 and 3n/2 for original fixed degree n
 *		Added Threads
 *		Modified ComputeLaterals/Branches to return Demand*
 *		 (not bad because it's contained in Lateral-Struct etc
 *		 anyway, so no malloc necessary)
 * 17-dec-96	Pass P and Q als parameters, no global stuff!!
 *		Added recursive top-level function "Compute_Feeder"
 *		 to be able to express parallelism better, relying
 *		 on newly added explicit feeder layer
 *		Instrumented Compute_Feeder
 * 22-dec-96	More flexible creation predicates (keep N etc)
 *		If threads_really_disabled defined -> all thread
 *		 stuff removed (for comparision)
 * 02-jan-97    Problem specific parameters for thread creation:
 *               depth_number_xxx, new_thread_p_depth_xxx and combination
 *              Ifdefs for use_depth_p and use_general_p
 *              defines for thread_{rd,wr,un}lock
 *		Use new_thread_predicate_XXX instead of new_thread_p now
 * 16-jul-97	No return value for recursive functions, use reference
 *		 call parameters instead (i.e. Demand *x() -> x(Demand **d))
 *		 
 */




/*----------------------------------------------------------------------*/
/* Leaf optimization procedures                 */

void optimize_node (double pi_R, double pi_I, double *P, double *Q);
double find_g (double P, double Q);
double find_h (double P, double Q);
double find_gradient_f (double pi_R, double pi_I, double* gradient,
			double P, double Q);
double find_gradient_g (double* gradient, double P, double Q);
double find_gradient_h (double* gradient, double P, double Q);
void find_dd_grad_f (double pi_R, double pi_I, double* dd_grad,
		     double P, double Q);
double make_orthogonal (double* v_mod, double* v_static);

void Compute_Feeder(Demand **d, Feeder f, double theta_R, double theta_I);
void Compute_Lateral(Demand **d, Lateral l, double theta_R, double theta_I, 
                       double pi_R, double pi_I);
void Compute_Branch(Demand **d, Branch br, double theta_R, double theta_I, 
                      double pi_R, double pi_I);
void Compute_Leaf(Demand **d, Leaf l, double pi_R, double pi_I);


void Compute_Tree(Root r) {
  Demand *a;
  double theta_R,theta_I;

  /* D.P = D.Q = 0 ?!?!?!? */
  
  theta_R = r->theta_R;
  theta_I = r->theta_I;

  Compute_Feeder(&a, r->feeder, theta_R, theta_I);

  r->D.P = a->P;
  r->D.Q = a->Q;
}



void Compute_Feeder(Demand **d, Feeder f,
                       double theta_R,
                       double theta_I)
{
    Demand *a1;
    Demand *a2;
    Feeder next = f->next_feeder;


    if (next != NULL) {

	Compute_Feeder(&a1, next, theta_R, theta_I);
    }

    Compute_Lateral(&a2, f->lateral,theta_R,theta_I,theta_R,theta_I);


    /* NEEDRESULTS */
    
    if (next != NULL) {
	f->D.P = a1->P + a2->P;
	f->D.Q = a1->Q + a2->Q;
    } else {
	f->D.P = a2->P;
	f->D.Q = a2->Q;
    }

    *d = &(f->D);	/* was: return &(f->D) */
}


void Compute_Lateral(Demand **d, Lateral l, double theta_R, double theta_I, 
                       double pi_R, double pi_I)
{
  Demand *a1;
  Demand *a2;
  double new_pi_R, new_pi_I;
  double a,b,c,root;
  Lateral next;
  Branch br;

  new_pi_R = pi_R + l->alpha*(theta_R+(theta_I*l->X)/l->R);
  new_pi_I = pi_I + l->beta*(theta_I+(theta_R*l->R)/l->X);

  next = l->next_lateral;
  if (next != NULL)  {
      Compute_Lateral(&a1, next,theta_R,theta_I,new_pi_R,new_pi_I);
  }

  br = l->branch;
  Compute_Branch(&a2, br,theta_R,theta_I,new_pi_R,new_pi_I);


  /* NEEDRESULTS */

  
  if (next != NULL) {
    l->D.P = a1->P + a2->P;
    l->D.Q = a1->Q + a2->Q;
  } else {
    l->D.P = a2->P;
    l->D.Q = a2->Q;
  }

  /* compute P,Q */
  a = l->R*l->R + l->X*l->X;  
  b = 2*l->R*l->X*l->D.Q - 2*l->X*l->X*l->D.P - l->R;
  c = l->R*l->D.Q - l->X*l->D.P;
  c = c*c + l->R*l->D.P;
  root = (-b-sqrt(b*b-4*a*c))/(2*a);
  l->D.Q = l->D.Q + ((root-l->D.P)*l->X)/l->R;
  l->D.P = root;

  /* compute alpha, beta */
  a = 2*l->R*l->D.P;
  b = 2*l->X*l->D.Q;
  l->alpha = a/(1-a-b);
  l->beta = b/(1-a-b);
  *d = &(l->D);
}

void Compute_Branch(Demand **d, Branch br, double theta_R, double theta_I, 
                      double pi_R, double pi_I){
  Demand *a1;
  Demand *a2;
  Demand  tmp;
  double new_pi_R, new_pi_I;
  double a,b,c,root;
  Leaf l;
  Branch next;
  int i;
  
  new_pi_R = pi_R + br->alpha*(theta_R+(theta_I*br->X)/br->R);
  new_pi_I = pi_I + br->beta*(theta_I+(theta_R*br->R)/br->X);

  next = br->next_branch;
  if (next != NULL)  {
      Compute_Branch(&a1, next,theta_R,theta_I,new_pi_R,new_pi_I);
  }

  /* Initialize tmp */
  tmp.P = 0.0; tmp.Q = 0.0;

  for (i=0; i<br->my_leaves_per_branch; i++) {
    l = br->leaves[i];
    Compute_Leaf(&a2, l,new_pi_R,new_pi_I);
    tmp.P += a2->P;
    tmp.Q += a2->Q;
  }

  /* NEEDRESULTS */

  if (next != NULL) {
    br->D.P = a1->P + tmp.P;
    br->D.Q = a1->Q + tmp.Q;
  } else {
    br->D.P = tmp.P;
    br->D.Q = tmp.Q;
  }

  /* compute P,Q */
  a = br->R*br->R + br->X*br->X;  
  b = 2*br->R*br->X*br->D.Q - 2*br->X*br->X*br->D.P - br->R;
  c = br->R*br->D.Q - br->X*br->D.P;
  c = c*c + br->R*br->D.P;
  root = (-b-sqrt(b*b-4*a*c))/(2*a);
  br->D.Q = br->D.Q + ((root-br->D.P)*br->X)/br->R;
  br->D.P = root;
  /* compute alpha, beta */
  a = 2*br->R*br->D.P;
  b = 2*br->X*br->D.Q;
  br->alpha = a/(1-a-b);
  br->beta = b/(1-a-b);

  *d = &(br->D);
}

void Compute_Leaf(Demand **d, Leaf l, double pi_R, double pi_I) {
  double P = l->D.P;
  double Q = l->D.Q;
  
  optimize_node(pi_R,pi_I,&P,&Q);

  if (P<0.0) {
    P = 0.0;
    Q = 0.0;
  }
  l->D.P = P;
  l->D.Q = Q;
  *d = &(l->D);
}

/*----------------------------------------------------------------------*/

void optimize_node (double pi_R, double pi_I, double *P, double *Q)
{
    double	    g;
    double	    h;

    double	    grad_f[2];
    double	    grad_g[2];
    double	    grad_h[2];
    double	    dd_grad_f[2];
    double	    magnitude;

    int		    i;
    double	    total;
    double	    max_dist;

    do {
	/* Move onto h=0 line */
	h=find_h (*P, *Q);
	if (fabs (h)>H_EPSILON) {
	    magnitude=find_gradient_h (grad_h, *P, *Q);
	    total=h/magnitude;
	    (*P)-=total*grad_h[0];
	    (*Q)-=total*grad_h[1];
	}

	/* Check that g is still valid */
	g=find_g (*P, *Q);
	if (g>G_EPSILON) {
	    magnitude=find_gradient_g (grad_g, *P, *Q);
	    find_gradient_h (grad_h, *P, *Q);
	    magnitude*=make_orthogonal (grad_g,grad_h);
	    total=g/magnitude;
	    (*P)-=total*grad_g[0];
	    (*Q)-=total*grad_g[1];
	}

	/* Maximize benefit */
	magnitude=find_gradient_f (pi_R,pi_I,grad_f, *P, *Q);
	find_dd_grad_f (pi_R,pi_I,dd_grad_f, *P, *Q);
	total=0.0;
	for (i=0; i<2; i++)
	    total+=grad_f[i]*dd_grad_f[i];
	magnitude/=fabs (total);
	find_gradient_h (grad_h, *P, *Q);
	magnitude*=make_orthogonal (grad_f,grad_h);
	find_gradient_g (grad_g, *P, *Q);
	total=0.0;
	for (i=0; i<2; i++)
	    total+=grad_f[i]*grad_g[i];
	if (total>0) {
	    max_dist=-find_g (*P, *Q)/total;
	    if (magnitude>max_dist)
		magnitude=max_dist;
	}
	(*P)+=magnitude*grad_f[0];
	(*Q)+=magnitude*grad_f[1];

	h=find_h (*P, *Q);
	g=find_g (*P, *Q);
	find_gradient_f (pi_R,pi_I,grad_f, *P, *Q);
	find_gradient_h (grad_h, *P, *Q);

    } while (fabs (h)>H_EPSILON || g>G_EPSILON ||
	    (fabs (g)>G_EPSILON &&
		fabs (grad_f[0]*grad_h[1]-grad_f[1]*grad_h[0])>F_EPSILON));
}

double find_g (double P, double Q)
{
    return (P*P+Q*Q-0.8);
}

double find_h (double P, double Q)
{
    return (P-5*Q);
}

double find_gradient_f (double pi_R, double pi_I, double* gradient,
			double P, double Q)
{
    int		    i;
    double	    magnitude=0.0;

    gradient[0]=1/(1+P)-pi_R;
    gradient[1]=1/(1+Q)-pi_I;
    for (i=0; i<2; i++)
	magnitude+=gradient[i]*gradient[i];
    magnitude=sqrt (magnitude);
    for (i=0; i<2; i++)
	gradient[i]/=magnitude;

    return magnitude;
}

double find_gradient_g (double* gradient, double P, double Q)
{
    int		    i;
    double	    magnitude=0.0;

    gradient[0]=2*P;
    gradient[1]=2*Q;
    for (i=0; i<2; i++)
	magnitude+=gradient[i]*gradient[i];
    magnitude=sqrt (magnitude);
    for (i=0; i<2; i++)
	gradient[i]/=magnitude;

    return magnitude;
}

double find_gradient_h (double* gradient, double P, double Q)
{
    int		    i;
    double	    magnitude=0.0;

    gradient[0]=1.0;
    gradient[1]=-5.0;
    for (i=0; i<2; i++)
	magnitude+=gradient[i]*gradient[i];
    magnitude=sqrt (magnitude);
    for (i=0; i<2; i++)
	gradient[i]/=magnitude;

    return magnitude;
}

void find_dd_grad_f (double pi_R, double pi_I, double* dd_grad,
		     double P, double Q)
{
    double	    P_plus_1_inv=1/(P+1);
    double	    Q_plus_1_inv=1/(Q+1);
    double	    P_grad_term=P_plus_1_inv-pi_R;
    double	    Q_grad_term=Q_plus_1_inv-pi_I;
    double	    grad_mag;
    
    grad_mag=sqrt (P_grad_term*P_grad_term+Q_grad_term*Q_grad_term);

    dd_grad[0]=-P_plus_1_inv*P_plus_1_inv*P_grad_term/grad_mag;
    dd_grad[1]=-Q_plus_1_inv*Q_plus_1_inv*Q_grad_term/grad_mag;
}

double make_orthogonal (double* v_mod, double* v_static)
{
    int		    i;
    double	    total=0.0;
    double	    length=0.0;

    for (i=0; i<2; i++)
	total+=v_mod[i]*v_static[i];
    for (i=0; i<2; i++) {
	v_mod[i]-=total*v_static[i];
	length+=v_mod[i]*v_mod[i];
    }
    length=sqrt (length);
    for (i=0; i<2; i++)
	v_mod[i]/=length;

    if (1-total*total<0)    /* Roundoff error--vectors are parallel */
	return 0;

    return sqrt (1-total*total);
}

/*----------------------------------------------------------------------*/


/* included main.c here */


/* For copyright information, see olden_v1.0/COPYRIGHT */

/* main.c
 *
 * By:  Martin C. Carlisle
 * 6/16/94
 *
 * main routine for the Power Pricing problem
 * adapted from code by:  Steve Lumetta, Sherry Li, and Ismail Khalil
 * University of California at Berkeley
 *
 *
 * Adapted by Stefan U. Haenssgen 15-dec-96
 *
 * 15-dec-96    Random degree varying from n/2..3n/2 for original fixed
 *               degree n, enlarged profile sizes accordingly
 *
 */


#include <sys/time.h>
#include <sys/timeb.h>



/* Domain of thetaR->P map is 0.65 to 1.00 [index*0.01+0.65] */
double map_P[36] =
{8752.218091048, 8446.106670416, 8107.990680283,
 7776.191574285, 7455.920518777, 7146.602181352,
 6847.709026813, 6558.734204024, 6279.213382291,
 6008.702199986, 5746.786181029, 5493.078256495,
 5247.206333097, 5008.828069358, 4777.615815166,
 4553.258735900, 4335.470002316, 4123.971545694,
 3918.501939675, 3718.817618538, 3524.683625800,
 3335.876573044, 3152.188635673, 2973.421417103,
 2799.382330486, 2629.892542617, 2464.782829705,
 2303.889031418, 2147.054385395, 1994.132771399,
 1844.985347313, 1699.475053321, 1557.474019598,
 1418.860479043, 1283.520126656, 1151.338004216};

#define      MIN_THETA_R       0.65
#define      PER_INDEX_R       0.01
#define      MAX_THETA_R       0.995

/* Domain of thetaI->Q map is 0.130 to 0.200 [index*0.002+0.130] */
double map_Q[36] =
{1768.846590190, 1706.229490046, 1637.253873079,
 1569.637451623, 1504.419525242, 1441.477913810,
 1380.700660446, 1321.980440476, 1265.218982201,
 1210.322424636, 1157.203306183, 1105.780028163,
 1055.974296746, 1007.714103979, 960.930643875,
 915.558722782, 871.538200178, 828.810882006,
 787.322098340, 747.020941334, 707.858376214,
 669.787829741, 632.765987756, 596.751545633,
 561.704466609, 527.587580585, 494.365739051,
 462.004890691, 430.472546686, 399.738429196,
 369.773787595, 340.550287137, 312.041496095,
 284.222260660, 257.068973074, 230.557938283};

#define      MIN_THETA_I       0.13
#define      PER_INDEX_I       0.002
#define      MAX_THETA_I       0.199

main(int argc,char *argv[])
{
  Root r;
  int    i,finished=0;
  double d_theta_R,d_theta_I;
  struct timeval starttime;
  struct timeval endtime;
  char   *ev;

                  
  dealwithargs(argc,argv);

  printf("power: %d feeders, %d laterals, %d branches, %d leaves\n",
       num_feeders,laterals_per_feeder,branches_per_lateral,leaves_per_branch);
  printf("(Degrees chosen randomly between n/2 and 3n/2)\n");
         

  gettimeofday(&starttime,NULL);

  /* initial pass */
  r = build_tree();
  chatting("Built tree\n");
  Compute_Tree(r);
  r->last.P = r->D.P;
  r->last.Q = r->D.Q;
  r->last_theta_R = r->theta_R;
  r->last_theta_I = r->theta_I;
  r->theta_R = TREE_THETA_R;
  r->theta_I = TREE_THETA_I;
  
  while (!finished) {
    Compute_Tree(r);
    chatting("TR=%13.9f, TI=%13.9f, P0=%13.9f, Q0=%13.9f\n",
           r->theta_R,r->theta_I,r->D.P,r->D.Q);
    if ((r->D.P == 0.0) && (r->D.Q == 0.0)) {
        printf("\n !!!  P = Q = 0.0 ... aborting!!!\n\n");
        finished = 1;
        exit(99);
    }
    if (fabs(r->D.P/10000.0 - r->theta_R) < ROOT_EPSILON &&
        fabs(r->D.Q/10000.0 - r->theta_I) < ROOT_EPSILON) {
      finished = 1;
    } else {
      i = (int)((r->theta_R - MIN_THETA_R) / PER_INDEX_R);
      if (i<0) i=0;
      if (i>35) i=35;
      d_theta_R = -(r->theta_R - r->D.P/10000.0) /
        (1 - (map_P[i+1] - map_P[i]) / (PER_INDEX_R * 10000.0));

      i = (int)((r->theta_I - MIN_THETA_I) / PER_INDEX_I);
      if (i<0) i=0;
      if (i>35) i=35;
      d_theta_I = -(r->theta_I - r->D.Q/10000.0) /
        (1 - (map_Q[i+1] - map_Q[i]) / (PER_INDEX_I * 10000.0));
 
      chatting("D TR-%13.9f, TI=%13.9f\n", d_theta_R,d_theta_I);
      r->last.P = r->D.P;
      r->last.Q = r->D.Q;
      r->last_theta_R = r->theta_R;
      r->last_theta_I = r->theta_I;
      r->theta_R = r->theta_R + d_theta_R;
      r->theta_I = r->theta_I + d_theta_I;
    }
  } /* while */
  
  gettimeofday(&endtime,NULL);

  chatting("Elapsed time %f\n",
           (endtime.tv_sec - starttime.tv_sec) +
           (endtime.tv_usec - starttime.tv_usec)/1000000.0 );

  
  exit(0);
  
}


/* For copyright information, see olden_v1.0/COPYRIGHT */

/* build.c
 *
 * By:  Martin C. Carlisle
 * 6/15/94
 * builds the tree for the Power Pricing problem
 *
 * based on code by:  Steve Lumetta, Sherry Li, and Ismail Khalil
 * University of California at Berkeley
 *
 *
 * Adapted by Stefan U. Haenssgen 15-dec-96
 *
 * 15-dec-96	Random degree varying from n/2..3n/2 for original fixed
 *		 degree n
 * 17-dec-96	Added explicit feeder layer, build_feeder()
 * 18-dec-96	Count number of leafs generated
 *		Fixed rnd generator start (reproducible results!!)
 * 29-dec-96	Loss factors etc as #defined parameters
 */

#include <stdlib.h>             /* Solaris, for rand() */
#include <sys/time.h>
#include <sys/timeb.h>

#undef DEBUG/**/
/*#define DEBUG 1/**/

extern int rand();
extern void srand();

/* Random value in the range of n/2..3n/2 */
#define myrand(n) ( (rand()) % (n) + (int)((n+1)/2) )

long numleafs;

/* NOPARALLEL */
/* NOSTATS */
Root build_tree() 
{
  register int i;
  register Root t;
  register Lateral l;
  int      nf;
  unsigned int seed;

  struct timeval now;
  
  gettimeofday(&now,NULL);
  /* seed = (unsigned int) (now.tv_sec+now.tv_usec); */
  seed = (unsigned int) 123;
  
  srand(seed);			/* init rnd generator */

  numleafs=0;
  t  = (Root) malloc(sizeof(*t));
  nf = myrand(num_feeders);
#ifdef DEBUG
  printf("seed = %d\n",seed);
  printf("num_feeders = %d (fixed = %d)\n",nf, num_feeders);
#endif
;  t->feeder = build_feeder(i*nf,nf);
  t->theta_R = TREE_THETA_R;
  t->theta_I = TREE_THETA_I;
  printf("%d leafs in tree\n",numleafs);
  return t;
}


/* NOPARALLEL */
/* NOSTATS */
Feeder build_feeder(int i, int num) /* RECURSIVE */
{
  register Feeder f;
  register Feeder next;
  register Lateral l;
  int nl;
 
  if (num == 0) return NULL;

  f = (Feeder) malloc(sizeof(*f));

  next = build_feeder(i,num-1); /* RECURSIVE */

  nl = myrand(laterals_per_feeder);
#ifdef DEBUG
  printf("  laterals_per_feeder[%d] = %d (fixed = %d)\n",
	 i, nl, laterals_per_feeder);
#endif
;  l = build_lateral(i*nl,nl);

  f->next_feeder = next;
  f->lateral = l;
  return f;
}


/* NOPARALLEL */
/* NOSTATS */
Lateral build_lateral(int i, int num) /* RECURSIVE */
{
  register int j,k;
  register Lateral l;
  register Branch b;
  register Lateral next;
  int nb;
 
  if (num == 0) return NULL;

  l = (Lateral) malloc(sizeof(*l));

  next = build_lateral(i,num-1); /* RECURSIVE */

  nb = myrand(branches_per_lateral);
#ifdef DEBUG
  printf("    branches_per_lateral[%d] = %d (fixed = %d)\n",
	 num, nb, branches_per_lateral);
#endif
;  b = build_branch(i*nb, (num-1)*nb, nb);

  l->next_lateral = next;
  l->branch = b;
  l->R = LATERAL_R;
  l->X = LATERAL_X;
  l->alpha = LATERAL_ALPHA;
  l->beta = LATERAL_BETA;
  return l;
}

/* NOPARALLEL */
/* NOSTATS */
Branch build_branch(int i, int j, int num) /* RECURSIVE */
{
  register Leaf l;
  register Branch b;
  int nl;

  if (num == 0) return NULL;
  /* allocate branch */
  b = (Branch) malloc(sizeof(*b));
  
  /* fill in children */
  b->next_branch= build_branch(i,j,num-1); /* RECURSIVE */

  nl = myrand(leaves_per_branch);
  b->my_leaves_per_branch = nl;
#ifdef DEBUG
  printf("      leaves_per_branch[%d] = %d (fixed = %d)\n",
	 num, nl, leaves_per_branch);
#endif
;  for (i=0; i<nl; i++) {
    l = build_leaf();
    b->leaves[i] = l;
  }
  
  /* fill in values */
  b->R = BRANCH_R;
  b->X = BRANCH_X;
  b->alpha = BRANCH_ALPHA;
  b->beta = BRANCH_BETA;
  return b;
}

/* NOPARALLEL */
/* NOSTATS */
Leaf build_leaf()
{
  register Leaf l;

  l = (Leaf) malloc(sizeof(*l));
  l->D.P = 1.0;
  l->D.Q = 1.0;
  numleafs++;
  return l;
}


/* For copyright information, see olden_v1.0/COPYRIGHT */



int mylog(int num)
{
  int j=0,k=1;
  
  while(k<num) { k*=2; j++; }
  return j;
} 

int dealwithargs(int argc, char *argv[])
{
  if (argc > 1)
      num_feeders = atoi(argv[1]);
  else
      num_feeders = NUM_FEEDERS;
  if (argc > 2)
      laterals_per_feeder = atoi(argv[2]);
  else
      laterals_per_feeder = LATERALS_PER_FEEDER;
  if (argc > 3)
      branches_per_lateral = atoi(argv[3]);
  else
      branches_per_lateral = BRANCHES_PER_LATERAL;
  if (argc > 4)
      leaves_per_branch = atoi(argv[4]);
  else
      leaves_per_branch = LEAVES_PER_BRANCH;

  
  __NumNodes = 1;
  __NDim = mylog(__NumNodes);

  return __NumNodes;
}
