/*
 * profile.c
 *
 * profiling functions
 *
 * Stefan U. Haenssgen  27-nov-96
 *
 *
 * 27-nov-96	1st version
 *		Global array for stats
 *		ProfileNewStats() for allocation & global initalization
 *		ProfileInitLine() for initialization of one-line's-statistics
 *		ProfilePrint() for output of complete profile
 *		ProfileFree() for freeing the whole structure
 *		Use blocks of structs instead of allocating the all
 *		 structs one by one (i.e. alloc all lines, then all depth
 *		 records for a line and all their degrees)
 *		Added documentation
 * 28-nov-96	Compute depth/degree sums on output
 *		Initial number of lines/depth/degree as parameter
 *		 for InitProfile
 *		Parameter "info" for line initialization (string)
 *		More comments
 * 29-nov-96	Print "[...]" also between non-zero lines on output
 * 17-jul-97	Use lines range 0..n, not 0..n-1 (consistently in
 *		 init and output)
 *
 *
 * TODO:	- dynamically grow arrays (depth, degree) when current size
 *		  is exceeded... involves check at runtime, expensive...
 *		- output relies on the maxdegree of all depths of a line
 *		  to be the same as the depth=0 maxdegree... merge the
 *		  degree into the depth record?
 *		- concurrent execution - either critical section around
 *		  the incrementing of the count (no extra postprocessing and
 *		  allocation, but expensive updates) or new statistics record
 *		  for each thread and merging of records (updates cheap but
 *		  allocation/merge phase)
 *		- measure how long each leaf/node takes to execute?
 */

#define profile_c

#include "profile.h"

profile_stat_t *profile_stat;		/* Default statistics collection */


/*
 * Allocate new statistics collection and set initial sizes of
 * lines, depths and degrees - useful if e.g. maximum recursion
 * degree is statically known.
 * Initialize all line statistics to "no data" - individual lines
 * are explicitly initialized for statistics collection using
 * ProfileInitLine()
 */
profile_stat_t *ProfileNewStats(int ilines, int idepth, int idegree)
{
    int i;
    profile_stat_t *stat = (profile_stat_t*) malloc(sizeof(profile_stat_t));
    
    stat->maxlines      = ilines;	/* Set initial sizes */
    stat->initialdepth  = idepth;
    stat->initialdegree = idegree;
    stat->lines         = (profile_line_t*) malloc ((ilines+1) *
						    sizeof(profile_line_t));

    for (i=0; i<=ilines; i++) {		/* Initialize empty lines */
	stat->lines[i].info	= (char*) NULL;
	stat->lines[i].maxdepth = 0;
	stat->lines[i].depths   = (profile_depth_t*) NULL;
    }
    return(stat);
}


/*
 * Create a new statistics collection with the default initial sizes
 */
profile_stat_t *ProfileNewDefaultStats(void)
{
    return ProfileNewStats(INITIALLINE, INITIALDEPTH, INITIALDEGREE);
}


/*
 * Initialize statistics collection for one line.
 * Copy "info" (containing any kind of annotation (procedure name, etc))
 * resp. ignore it if it is NULL.
 * Set depts and their corresponding degrees to initial sizes and
 * initialize with 0s.
 */
void ProfileInitLine(profile_stat_t *stat, int line, char *info)
{
    int i,j;
    
    if (stat->lines[line].maxdepth) {
	printf("ProfileInitLine: stats for line %d already initialized!\n",
	       line);
	exit(110);
    }

    /* Copy information string and initialize depth/degree records
     * with default sizes */
    
    if (info)
	stat->lines[line].info = (char*)strdup(info);
    stat->lines[line].maxdepth = stat->initialdepth;
    stat->lines[line].depths   = (profile_depth_t*) malloc(stat->initialdepth *
				 		     sizeof(profile_depth_t));

    for (i=0; i<stat->initialdepth; i++) {
	stat->lines[line].depths[i].maxdegree = stat->initialdegree;
	stat->lines[line].depths[i].degrees = (long*)malloc(
                                          stat->initialdegree * sizeof(long));
	for (j=0; j<stat->initialdegree; j++)
	    stat->lines[line].depths[i].degrees[j] = (long)0;
    }

}

/*
 * Output the profile for the used range of lines. Compute sums of
 * all degrees and depths of each line entry.
 */
void ProfilePrint(profile_stat_t *stat)
{
    int  i,j,k;
    char lastwasempty;

    printf("\nProfile:\n");

    lastwasempty=0;
    
    for (i=0; i<=stat->maxlines; i++) {		/* Check each line entry */
	if (stat->lines[i].maxdepth) {		/* Non-Empty -> output it */
	    
	    long *degreesums;			/* Sum of all degree-columns */
	    long  depthsum;			/* Sum of one depth-row */
	    long  count;

	    lastwasempty = 0;
	    
	    degreesums = (long*) malloc (stat->lines[i].depths[0].maxdegree *
				   sizeof(long));
	    for (j=0; j<stat->lines[i].depths[0].maxdegree; j++)
		degreesums[j] = 0;
	    
	    /* Output one depth line with all its degree-counts */

	    printf("%5d - %s\n",i,
		   (stat->lines[i].info?stat->lines[i].info:"(no info)"));
	    
	    for (j=0; j<stat->lines[i].maxdepth; j++) {
		
		depthsum = 0;
		printf("        (%3d: ", j);
		
		for (k=0; k<stat->lines[i].depths[j].maxdegree; k++) {
		    count = stat->lines[i].depths[j].degrees[k];
		    printf("%4d ", count);
		    depthsum += count;
		    degreesums[k] += count;	/* Remember sum of degrees */
		}
		printf(" sum = %4d)\n", depthsum);
	    }
	    
	    /* Output sum of degrees over all depths */
	    
	    printf("        (sum: ");
	    depthsum = 0;
	    for (k=0; k<stat->lines[i].depths[0].maxdegree; k++) {
		printf("%4d ", degreesums[k]);
		depthsum += degreesums[k];
	    }
	    printf(" SUM = %4d)\n", depthsum);
		
	    free(degreesums);
	    
	} else {
	    
	    if (!lastwasempty)
		printf("[...]\n",i);		/* Nothing to output... */
	    lastwasempty=1;			/* only print once in a row */
	}
    }

}

/*
 * Free the given profile structure
 */
void ProfileFree(profile_stat_t *stat)
{
    int i,j,k;

    for (i=0; i<stat->maxlines; i++) {		/* Deallocate all ... */
	if (stat->lines[i].maxdepth) {
	    for (j=0; j<stat->lines[i].maxdepth; j++) 
		free(stat->lines[i].depths[j].degrees);
	    free(stat->lines[i].depths);
	    free(stat->lines);
	}
    }
    free(stat);
}


/*
 * Small test procedure - initalize stats collection, enter values, output them
 */
ProfileTest()
{
    profile_stat_t *stat = ProfileNewStats(10,5,5);
    ProfileInitLine(stat,1, "Line 1");
    ProfileInitLine(stat,5, "Line 5");
    ProfileInitLine(stat,9, (char*)NULL); /* test for empty info */
    ProfileAddStat(stat,1,2,3,10); /* line 1 depth 2 degree 3 cnt 10 */
    ProfileAddStat(stat,1,2,3,4);
    ProfileAddStat(stat,5,1,0,1);
    ProfileAddStat(stat,9,1,1,4);

    ProfilePrint(stat);

    ProfileFree(stat);
}

/*main(){ProfileTest();}*/ /* de-comment this for testing */
