/*
 * evaluate_profile.c
 *
 * Parses input from profile generation and evaluates it according
 * to an implicit "worst case tree" (how large could the largest
 * subtree be, for each depth)
 * Profiles end with a "[...]" line. If this line is missing between
 * two profiles, they are merged, i.e. their values are added up
 * cell by cell (e.g. for merging the profiles of two rcursive procedures
 * where a calls a and b, b just calls b (hierarchy levels))
 *
 *
 * Usage:   evaluate_profile [options] < profile
 *	    Options:
 *	    -v		Verbose output
 *	    -a		Evaluate average subtrees
 *	    -l		Evaluate largest possuble subtree (default)
 *	    -c C	Recommend strategy for C CPUs (default = 4)
 *
 * Stefan Haenssgen  08-jan-98
 *
 *
 * 08-jan-98	1st version based on tree_from_profile.c
 * 09-jan-98	Added option -c C  for number of CPUs
 *		Recommend Depth if largest subtree <= 1/cpunum 
 *		Also output procedure name & line
 *		Start profile parsing after end of first profile is read
 *		Recognize trees with not enough leafs (e.g. if lower
 *		 hierarchy is missing from tree) and output warning
 *		Don't recommend layers with less nodes than CPUs
 *		Allow for profile merging, don't reallocate cells
 * 		 if they already exist. Notify user of merging
 *		Recognize number of iterations (nodes at depth 0)
 *		 and calculate with fractions if they are >1
 *		 (special analyze_iterations procedure)
 * 14-jan-98	Also compute average subtree
 *		Output warning if number of threads > 3000 (i.e. number
 *		 of nodes above and including the start depth)
 * 15-jan-98	Recommend use of NOTHREAD annotation if degree=0/n
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define MAXDEGREE 20
#define MAXDEPTH  90

#define MAXTHREADS 3000		/* Thread number not to be exceeded	*/

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define TEXTCHUNKSIZE 16384 /* Size of text string chunks (4 allocation)*/

long *profiletab[MAXDEPTH];	/* Parsed profile table			*/
int   maxdegree = 0;		/* Maximum degree/depth encountered	*/
int   maxdepth  = 0;
char  procname[BUFSIZ];		/* Name&line# of the procedure profiled	*/
int   procline;
int   iterations = 0;

char  verbose   = FALSE;	/* Output flag				*/

#define LPS_MODE 1
#define AS_MODE  2

char treemode = LPS_MODE;	/* Evaluate largest possible or average?*/


/* Parse one profile line for degree counts, also return its depth
 * Input = profile line
 *    (  0:    0    0    0    0    0    0    0    0    1     0  sum =    1)
 *   or
 *    (sum: 204886  0    0    0    0    0    0    0 22765    0  SUM = 227651)
 * Profile entry ends with "[...]", followed possibly by the next
 * profile (which is ignored)
 */

#define M_NONE    0	/* States: before '('				*/
#define M_DEPTH   1	/* After '(' before ':' reading depth		*/
#define M_DEGREES 2	/* after ':' before sum/SUM reading degrees	*/
#define M_SUM     3	/* after sum/SUM reading sum			*/

void get_degree_counts(char *inputbuf, long *degreecounts, int *depth) {
    char *s;
    char  lastc;
    int   i;
    long  n;
    char  mode;
    char  nonnull;
    
    *depth = -1;			/* Empty read -> depth=-1	*/
    for (i=0; i<MAXDEGREE; i++)
	degreecounts[i] = 0;
    s = inputbuf;
    mode = M_NONE;
    lastc = (char) 0;
    nonnull = FALSE;

    while (*s) {
	if (*s == '(') {		/* Read depth now		*/
	    mode = M_DEPTH;
	    n = 0;
	} else if (*s == ':') {		/* Read degrees now		*/
	    *depth = n;
	    mode = M_DEGREES;
	    i = 0;
	    n = 0;
	} else if ((*s == 's') ||	/* Ignore sum ...		*/
		   (*s == 'S')) {
	    if ((nonnull) && (*depth > maxdepth))
	      maxdepth = *depth;	/* Remember max. non-null depth	*/
	    return;
	} else if ((*s >= '0') &&	/* Parse number			*/
		   (*s <= '9')) {
	    n = 10*n + (long)(*s - '0');
	} else if (*s == ' '){
	    if ((mode == M_DEGREES) &&
		(lastc >= '0') && (lastc <= '9')) {
		degreecounts[i] = n;
		if (n>0) {		/* Any non-null entries found?	*/
		  nonnull = TRUE;
		  if (i>maxdegree)	/* Keep track of max.degree	*/
		    maxdegree=i;
		}
		n = 0;			/* Space separates degree counts*/
		i++;			/* Last count read, now next one*/
	    }
	}
	lastc = *s;
	s++;
    }
    
    if ((nonnull) && (*depth > maxdepth))
      maxdepth = *depth;		/* Remember max. non-null depth	*/

}


/* Parse global input text (number of times a branch of each degree 0..x
 * was taken) and build internal array from it.
 */
void parse_profile() {
  char     inputbuf[TEXTCHUNKSIZE];
  int      inputlen = 0;
  int      depth;
  int      deg;
  long    *profileline;
  int      mergeflag = 0;

  maxdegree = 0;
  maxdepth  = 0;

  for (depth=0; depth < MAXDEPTH; depth++)
    profiletab[depth] = (long*) NULL;

  /* Read profile from stdin */

  if (gets(inputbuf) == NULL) {
      printf("Error reading first line!\n");
      exit(110);
  }

  sscanf(inputbuf, "%d - %s", &procline, procname);/* Read procedure info */
  
  /* Now parse the remaining profile:
   * Read line and branching degree counts.
   */

  while (gets(inputbuf)) {

      profileline = malloc(MAXDEGREE * sizeof(long));

      if (inputbuf[0] == '[') {	     /* Stop parsing after 1st profile! */
	break;
      }

      get_degree_counts(inputbuf, profileline, &depth);

      if (depth>=0) { 			/* Useful information read?	*/

	/* If no such line already existed, take the newly read
	 * line for the profile entry.
	 * If the line already has values, merge the two profiles
	 * by adding the new values to the existing ones
	 */
	if (profiletab[depth] == (long*)NULL) {
	  profiletab[depth] = profileline;    /* Remember profile line!	*/
	} else {
	  int deg;
	  for (deg=0; deg<MAXDEGREE; deg++) {
	    profiletab[depth][deg] = profiletab[depth][deg] +  
	                             profileline[deg];
	  }
	  free(profileline);
	  if (depth==0)			/* Start of a profile?		*/
	    mergeflag++;	  	/* Another profile merged	*/
	}
      }
  }

  if (mergeflag) {
    printf("note - merged %d profiles together!\n", mergeflag+1);
  }

  iterations = 0;
  for (deg=1; deg<=maxdegree; deg++)
    iterations = iterations + profiletab[0][deg];

}


void print_profile()
{
  int dep,deg;

  for (dep=0; dep<=maxdepth; dep++) {
    printf("%2d :", dep);
    for (deg=0; deg<=maxdegree; deg++) {
      printf(" %6d ",profiletab[dep][deg]);
    }
    printf("\n");
  }
}


/* Counts how many nodes are in the larges possible subtree
 * starting at the given depth. Give recommendations for depths
 * that are probably good for the parallelization on the given
 * number of CPUs.
 */
void analyze(int startdepth, int cpunum)
{
  int 	 i;
  int    dep, deg;
  long   nodes, leafs;
  long   grabnodes;
  long   allnodes, allleafs;
  long   thisdepthnodes;
  long   stilltograb;
  float	 allpercent;
  float  cpupercent = 100.0 / (float) cpunum;	/* Percentage of 1 CPU	*/

  nodes=1;			/* Keep track of nodes/leafs so far	*/
  leafs=0;
  grabnodes=1;			/* Start with one node as subtree	*/


  allnodes = 0;			/* Count how many nodes/leafs are above	*/
  allleafs = 0;
  for (dep=0; dep<startdepth; dep++) {
    for (deg=1; deg<=maxdegree; deg++)
      allnodes = allnodes + profiletab[dep][deg];
    allleafs = allleafs + profiletab[dep][0];
  }
  thisdepthnodes=0;
  for (deg=1; deg<=maxdegree; deg++)
    thisdepthnodes=thisdepthnodes+profiletab[startdepth][deg];
  if (allnodes+thisdepthnodes > MAXTHREADS) {
    printf("Depth %d NOT recommended, maximum thread number (%d) exceeded by %d!\n",
	   startdepth, MAXTHREADS, allnodes+thisdepthnodes-MAXTHREADS);
    return;
  }
  if (verbose) {
    printf("There are %5d nodes above and including depth %2d\n",
	   allnodes+thisdepthnodes, startdepth);
  }

  /* Construct "largest possible subtree":
   * Descend in profile starting from depth 1, always adding the
   * maximum possible number of nodes to the current "tree"
   * starting with the highest branching degree available.
   * No subtree can get larger than this one!
   */
  for (dep=startdepth; dep<=maxdepth; dep++) {
    long layernodes, layerleafs;

    stilltograb = grabnodes;

    /* Keep track of how many of the available nodes in this
     * layer resp. globally are consumed
     */
    layernodes=0;
    for (deg=1; deg<=maxdegree; deg++)
      layernodes=layernodes+profiletab[dep][deg];
    layerleafs = profiletab[dep][0];
    allnodes = allnodes + layernodes;
    allleafs = allleafs + layerleafs;

    if (verbose) {
      printf("\n");
      printf("depth %2d: %5d nodes and %5d leafs so far.\n",
	     dep, nodes, leafs);
      printf("that's %6.2f%% of the population up to this layer\n",
	     100.0*(float)stilltograb/(float)(allnodes+allleafs));
      printf("grabbing %5d nodes out of %5d nodes and %5d leafs here...\n",
	     stilltograb, layernodes, layerleafs);
      printf("that's %6.2f%% of this layer's population (nodes+leafs)\n",
	     100.0*(float)stilltograb/(float)(layernodes+layerleafs));
    }

    /* Now grab "stilltograb" nodes/leafs
     */
    grabnodes = 0;
    for (deg=maxdegree; ((stilltograb>0) && (deg>=0)); deg--) {
      long p = profiletab[dep][deg];
      long g = 0;

      /* Any node/leaf here at all? No -> go on to next lower degree
       */
      if (p) {
	if (p>=stilltograb) {	/* Enough nodes for me?			*/
	  g = stilltograb;
	  stilltograb = 0;	/* OK, satisfied now			*/
	} else {		/* Not enough, get all available here	*/
	  g = p;
	  stilltograb = stilltograb - p;
	}
	if (verbose) {
	  printf("  grabbed %5d nodes of degree %2d\n", g, deg);
	}

	if (deg>0)		/* Keep track of nodes/leafs in all	*/
	  nodes = nodes + g;
	else
	  leafs = leafs + g;

	/* Compute number of nodes to grab in the next depth layer,
	 * degree times the number of nodes grabbed right now
	 */
	grabnodes = grabnodes + deg * g;
      }
    }
  }
  
  if (stilltograb>0) {
    printf("warning, %d nodes left to grab but no more available!\n",
	   stilltograb);
    printf("is this a complete tree or are there leafs in lower hierarchies missing?\n");
  }


  allpercent = 100.0*(float)(nodes+leafs)/(float)(allnodes+allleafs);

  printf("\n");
  printf("all in all, the largest subtree has %5d nodes and %5d leafs.\n",
	   nodes, leafs);
  printf("that's %6.2f%% of the tree's nodes,\n",
	   100.0*(float)nodes/(float)(allnodes));
  printf("       %6.2f%% of its leafs,\n",
	   100.0*(float)leafs/(float)(allleafs));
  printf("   and %6.2f%% of its nodes+leafs\n", allpercent);
  if (allpercent <= cpupercent) {
    if (thisdepthnodes >= cpunum) {
      printf("-> RECOMMENDED: depth %2d for %2d CPUs, largest subtree is only %6.2f%%\n",
	     startdepth, cpunum, allpercent);
    } else {
      printf("-> not recommended, subtree size OK but layer has less than %d nodes (%d)!\n",
	     cpunum, thisdepthnodes);
    }
	     
  } else {
    printf("-> not recommended for %d CPUs, largest subtree larger than 1/%d (%6.2f%%)\n",
	   cpunum, cpunum, allpercent);
  }

}


/* Same as analyze but for more than one iteration - uses
 * fractions to approximate the impact of one iteration.
 * (just divides every value by the iteration count)
 */
void analyze_iterations(int startdepth, int cpunum)
{
  int 	 i;
  int    dep, deg;
  double nodes, leafs;
  double grabnodes;
  double allnodes, allleafs;
  double thisdepthnodes;
  double stilltograb;
  float	 allpercent;
  float  cpupercent = 100.0 / (float) cpunum;	/* Percentage of 1 CPU	*/

  nodes=1.0;			/* Keep track of nodes/leafs so far	*/
  leafs=0.0;
  grabnodes=1.0;		/* Start with one node as subtree	*/


  allnodes = 0.0;		/* Count how many nodes/leafs are above	*/
  allleafs = 0.0;
  for (dep=0; dep<startdepth; dep++) {
    for (deg=1; deg<=maxdegree; deg++)
      allnodes = allnodes + (double)profiletab[dep][deg]/(double)iterations;
    allleafs = allleafs + (double)profiletab[dep][0]/(double)iterations;
  }
  thisdepthnodes=0.0;
  for (deg=1; deg<=maxdegree; deg++)
    thisdepthnodes=thisdepthnodes+(double)profiletab[startdepth][deg]/
                                  (double)iterations;
  if (allnodes+thisdepthnodes > MAXTHREADS) {
    printf("Depth %d NOT recommended, maximum thread number (%d) exceeded by %d!\n",
	   startdepth, MAXTHREADS, (long)(allnodes+thisdepthnodes)-MAXTHREADS);
    return;
  }
  if (verbose) {
    printf("There are %5.0f nodes above and including depth %2d\n",
	   allnodes+thisdepthnodes, startdepth);
  }


  /* Construct "largest possible subtree":
   * Descend in profile starting from depth 1, always adding the
   * maximum possible number of nodes to the current "tree"
   * starting with the highest branching degree available.
   * No subtree can get larger than this one!
   */
  for (dep=startdepth; dep<=maxdepth; dep++) {
    double layernodes, layerleafs;

    stilltograb = grabnodes;

    /* Keep track of how many of the available nodes in this
     * layer resp. globally are consumed.
     * Calculate with fractions, taking the number of iterations
     * into account!
     */
    layernodes=0.0;
    for (deg=1; deg<=maxdegree; deg++)
      layernodes=layernodes+(double)profiletab[dep][deg]/(double)iterations;
    layerleafs = (double)profiletab[dep][0]/(double)iterations;
    allnodes = allnodes + layernodes;
    allleafs = allleafs + layerleafs;

    if (verbose) {
      printf("\n");
      printf("depth %2d: %7.1f nodes and %7.1f leafs so far.\n",
	     dep, nodes, leafs);
      printf("that's %6.2f%% of the population up to this layer\n",
	     100.0*(float)stilltograb/(float)(allnodes+allleafs));
      printf("grabbing %7.1f nodes out of %7.1f nodes and %7.1f leafs here...\n",
	     stilltograb, layernodes, layerleafs);
      printf("that's %6.2f%% of this layer's population (nodes+leafs)\n",
	     100.0*(float)stilltograb/(float)(layernodes+layerleafs));
    }

    /* Now grab "stilltograb" nodes/leafs
     */
    grabnodes = 0.0;
    for (deg=maxdegree; ((stilltograb>0.0) && (deg>=0)); deg--) {
      double p = (double) profiletab[dep][deg] / (double)iterations;
      double g = 0.0;

      /* Any node/leaf here at all? No -> go on to next lower degree
       */
      if (p>0.0) {
	if (p>=stilltograb) {	/* Enough nodes for me?			*/
	  g = stilltograb;
	  stilltograb = 0.0;	/* OK, satisfied now			*/
	} else {		/* Not enough, get all available here	*/
	  g = p;
	  stilltograb = stilltograb - p;
	}
	if (verbose) {
	  printf("  grabbed %7.1f nodes of degree %2d\n", g, deg);
	}

	if (deg>0)		/* Keep track of nodes/leafs in all	*/
	  nodes = nodes + g;
	else
	  leafs = leafs + g;

	/* Compute number of nodes to grab in the next depth layer,
	 * degree times the number of nodes grabbed right now
	 */
	grabnodes = grabnodes + deg * g;
      }
    }
  }
  
  if (stilltograb>0.0) {
    printf("warning, %.1f nodes left to grab but no more available!\n",
	   stilltograb);
    printf("is this a complete tree or are there leafs in lower hierarchies missing?\n");
  }

  allpercent = 100.0*(float)(nodes+leafs)/(float)(allnodes+allleafs);

  printf("\n");
  printf("all in all, the largest subtree has %5d nodes and %5d leafs.\n",
	   nodes, leafs);
  printf("that's %6.2f%% of the tree's nodes,\n",
	   100.0*(float)nodes/(float)(allnodes));
  printf("       %6.2f%% of its leafs,\n",
	   100.0*(float)leafs/(float)(allleafs));
  printf("   and %6.2f%% of its nodes+leafs\n", allpercent);
  if (allpercent <= cpupercent) {
    if (thisdepthnodes >= cpunum) {
      printf("-> RECOMMENDED: depth %2d for %2d CPUs, largest subtree is only %6.2f%%\n",
	     startdepth, cpunum, allpercent);
    } else {
      printf("-> not recommended, subtree size OK but layer has less than %d nodes (%.1f)!\n",
	     cpunum, thisdepthnodes);
    }
	     
  } else {
    printf("-> not recommended for %d CPUs, largest subtree larger than 1/%d (%6.2f%%)\n",
	   cpunum, cpunum, allpercent);
  }

}


/* Same as analyze iterations, but computes average subtree.
 * Doesn't differenciate between nodes and leafs (what would a
 * node of average degree 0.15 be anyway :)
 * Can also be used for just one iteration, uses fractions anyway.
 */
void analyze_average(int startdepth, int cpunum)
{
  int 	 i;
  int    dep, deg;
  double nodes;
  double grabnodes;
  double allnodes;
  double avgdegree;
  double thisdepthnodes;
  double stilltograb;
  float	 allpercent;
  float  cpupercent = 100.0 / (float) cpunum;	/* Percentage of 1 CPU	*/

  nodes=1.0;			/* Keep track of nodes so far		*/
  grabnodes=1.0;		/* Start with one node as subtree	*/


  allnodes = 0.0;		/* Count how many nodes are above	*/

  for (dep=0; dep<startdepth; dep++) {
    for (deg=0; deg<=maxdegree; deg++)
	allnodes = allnodes + (double)profiletab[dep][deg]/(double)iterations;
  }
  thisdepthnodes=0.0;
  for (deg=0; deg<=maxdegree; deg++)
    thisdepthnodes=thisdepthnodes+(double)profiletab[startdepth][deg]/
                                  (double)iterations;
  if (allnodes+thisdepthnodes > MAXTHREADS) {
    printf("Depth %d NOT recommended, maximum thread number (%d) exceeded by %d!\n",
	   startdepth, MAXTHREADS, (long)(allnodes+thisdepthnodes)-MAXTHREADS);
    return;
  }
  if (verbose) {
    printf("There are %5.0f nodes above and including depth %2d \n",
	   allnodes+thisdepthnodes, startdepth);
  }


  /* Construct "average subtree":
   * Descend in profile starting from depth 1, always adding the
   * average number of nodes to the current "tree"
   * using the average (weighted) branching degree.
   */
  for (dep=startdepth; dep<=maxdepth; dep++) {
    double layernodes;
    double grabbed;

    stilltograb = grabnodes;

    /* Keep track of how many of the available nodes in this
     * layer resp. globally are consumed.
     * Calculate with fractions, taking the number of iterations
     * into account!
     */
    layernodes=0.0;
    for (deg=0; deg<=maxdegree; deg++)
      layernodes=layernodes+(double)profiletab[dep][deg]/(double)iterations;
    allnodes = allnodes + layernodes;

    if (verbose) {
      printf("\n");
      printf("depth %2d: %7.1f nodes so far.\n",
	     dep, nodes);
      printf("that's %6.2f%% of the population up to this layer\n",
	     100.0*stilltograb/allnodes);
      printf("grabbing %7.1f nodes out of %7.1f nodes here...\n",
	     stilltograb, layernodes);
      printf("that's %6.2f%% of this layer's population\n",
	     100.0*stilltograb/layernodes);
    }

    /* Now grab "stilltograb" nodes/leafs
     */
    grabnodes = 0.0;
#if 1	/* Arithmetisches Mittel */
    avgdegree = 0.0;
    for (deg=0; deg<=maxdegree; deg++) {
        double p  = (double) profiletab[dep][deg] / (double)iterations;
	double d  = (double) deg;
	avgdegree = avgdegree + p * d;
    }
    avgdegree = avgdegree / layernodes;
#endif
#if 0	/* Quadratisches Mittel */
    avgdegree = 0.0;
    for (deg=0; deg<=maxdegree; deg++) {
        double p  = (double) profiletab[dep][deg] / (double)iterations;
	double d  = (double) deg;
	avgdegree = avgdegree + p * d*d;
    }
    avgdegree = sqrt(1.0/layernodes * avgdegree);
#endif
#if 0	/* Geometrisches Mittel */
    avgdegree = 1.0;
    for (deg=0; deg<=maxdegree; deg++) {
        double p  = (double) profiletab[dep][deg] / (double)iterations;
	double d  = (double) deg;
	avgdegree = avgdegree * pow(d, p);
    }
    avgdegree = pow(avgdegree, 1.0/layernodes);
#endif
    if (layernodes>=stilltograb) {	/* Enough nodes for me?		*/
	grabbed = stilltograb;
	stilltograb = 0.0;		/* OK, satisfied now		*/
    } else {			/* Not enough, get all available here	*/
	grabbed = layernodes;
	stilltograb = stilltograb - layernodes;
    }
    if (verbose) {
	printf("  grabbed %7.1f nodes of average degree %5.1f\n",
	       grabbed, avgdegree);
    }
    nodes = nodes + grabbed;

    /* Compute number of nodes to grab in the next depth layer,
     * degree times the number of nodes grabbed right now
     */
    grabnodes = grabbed * avgdegree;
  }
  
  if (stilltograb>0.0) {
    printf("warning, %.1f nodes left to grab but no more available!\n",
	   stilltograb);
    printf("is this a complete tree or are there leafs in lower hierarchies missing?\n");
  }

  allpercent = 100.0*(float)(nodes)/(float)(allnodes);

  printf("\n");
  printf("all in all, the average subtree has %.1f nodes.\n",
	   nodes);
  printf("that's %6.2f%% of the tree's nodes\n",
	   100.0*(float)nodes/(float)(allnodes));
  if (allpercent <= cpupercent) {
    if (thisdepthnodes >= cpunum) {
      printf("-> RECOMMENDED: depth %2d for %2d CPUs, average subtree is only %6.2f%%\n",
	     startdepth, cpunum, allpercent);
    } else {
      printf("-> not recommended, subtree size OK but layer has less than %d nodes (%.1f)!\n",
	     cpunum, thisdepthnodes);
    }
	     
  } else {
    printf("-> not recommended for %d CPUs, average subtree larger than 1/%d (%6.2f%%)\n",
	   cpunum, cpunum, allpercent);
  }

}


/* Recommends using the NOTHREAD annotation if the profile's branching
 * degree is always 0 or N and nothing else
 */
void recommend_nothread() {
    int	 n;
    int  dep, deg;
    char alwaysn;

    n       = 0;
    alwaysn = FALSE;
    
    for (dep=0; dep<=maxdepth; dep++) {
	for (deg=1; deg<=maxdegree; deg++) {	/* Only check nodes:	*/
	    if (profiletab[dep][deg] != 0) {	/* Non-zero degree?	*/
		if (n == 0) {			/* N not set yet?	*/
		    n = deg;			/* Remember degree!	*/
		    alwaysn = TRUE;
		} else {			/* N set, is it same?	*/
		    if (n != deg) {		/* No -> not always..	*/
			alwaysn = FALSE;
		    }
		}
	    }
	}
    }
    if ((alwaysn) && (n)) {
	printf("recommend using the NOTHREAD annotation for branching degree %d\n",
	       n);
	printf(" since degree is always 0 or %d.\n", n);
    }
    
}

main(int argc, char **argv)
{
  int     i;
  int	  d;
  int     cpunum = 4;		/* Number of CPUs for recommendations	*/
  

  /* Parse command line */
      
  for (i=1; i<argc; i++) {
      if (!strcmp(argv[i],"-v")) {
	  printf("Verbose output\n");
	  verbose = TRUE;
      } else if (!strcmp(argv[i],"-a")) {
	  treemode = AS_MODE;
      } else if (!strcmp(argv[i],"-l")) {
	  treemode = LPS_MODE;
      } else if (!strcmp(argv[i],"-c")) {
          if (i==argc-1) {
              printf("Error, -c needs argument (number of processors)\n");
              exit(22);
          }
          cpunum = atoi(argv[i+1]);
          i++;
      } else {
	  printf("%s - parse profile from stdin and computes the larges possible subtrees\n",
		 argv[0]);
	  printf("Usage:   %s [options] < profile\n", argv[0]);
          printf("	    Options:\n");
          printf("	    -v		Verbose output\n");
	  printf("	    -a		Evaluate average subtrees\n");
	  printf("	    -l		Evaluate largest possuble subtree (default)\n");
          printf("	    -c C	Recommend strategy for C CPUs (default = 4)\n");
	  exit(123);
      }
  }

  parse_profile();

  printf("profiled procedure '%s' at source line %d\n", procname, procline);
  printf("maximum degree = %d, maximum depth = %d\n", maxdegree, maxdepth);
  printf("profile covers %d iterations\n", iterations);

  if (verbose) {
    print_profile();
  }

  recommend_nothread();		/* Check if degree is 0 or N -> NOTHREAD OK! */
  
  if (treemode == LPS_MODE) {
      printf("performing Largest Possible Subtree");
  } else {
      printf("performing Average Subtree");
  }
  printf(" analysis for %d CPUs\n", cpunum);

  for (d=1; d<maxdepth; d++) {
    printf("\nanalyzing depth %d:\n", d);
    if (treemode == AS_MODE) {		/* Average always uses fractions  */
	analyze_average(d, cpunum);
    } else {
	if (iterations == 1) {
	    analyze(d, cpunum);		/* LPS: 1 iteration -> exact ints */
	} else {			/* >1 iteration -> fractions	  */
	    analyze_iterations(d, cpunum);
	}
    }
  }

  exit(0);
}

