/*
 * tree_graph_output.c
 *
 * Parses tree input from statistics generation and generates
 * in-memory tree. Outputs this tree in comdraw or pstricks format
 * on stdout.
 *
 *
 * Usage:   tree_graph_output [options] < treefile
 *	    options:
 *	    -comdraw	Output in comdraw format (also '-c')
 *	    -pstricks	Output in pstricks format (also '-p', default)
 *	    -x n	Sets x spacing of nodes to n (default=20)
 *	    -y n	Sets y spacing of nodes to n (default=12)
 *	    -s n	Sets node diameter to n      (default= 2)
 *	    treefile	ASCII representation of recursion tree, e.g.
 *			"(()())" for simple binary tree, may be annotated
 *		        with procedure numbers and subtree timings 
 *
 *
 * Stefan Haenssgen  06-nov-97
 *
 * 06-nov-97	1st version based on tree_parser
 * 07-nov-97	X/Y spacing and node size as parameters
 * 12-nov-97	Output in comdraw or pstricks (new) format
 *		General line/node drawing and init/end procedures
 *		tree_size for determining the size of the drawing
 *
 *
 * ??? output tree during parsing, no intermediate tree in memory?
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define TEXTCHUNKSIZE 16384 /* Size of text string chuncs (4 allocation)*/

#define CHILDNUM 10	/* Default starting number of children for nodes*/

typedef struct _node *nodeptr;

typedef struct _node {
  int      numchildren;	/* Number of children this node has		*/
  nodeptr *children;	/* Array of children for this node		*/
  int 	   numnodes;	/* Number of nodes (without leafs) in subtree	*/
  int      numleafs;	/* Number of leafs in this subtree		*/
  int      procnum;	/* Number code for the node's procedure		*/
  long     cputime;	/* Used CPU time for this subtree in millisec.	*/
  int	  *procnodes;	/* Number of leafs for each procedure type	*/
  int	  *procleafs;	/* Ditto for number of nodes (from 0..maxpn-1)	*/
  int     *info;	/* Any additional information (size, ...)	*/
} node;

#define NULLPTR (nodeptr)NULL;

char *text;		/* Global pointer to current input position	*/

int maxprocnum;		/* Global maximum number of proc-indices found	*/
                        /* Zero here means no proc-numbers were found   */
                        /* at all. procnums range 1..mp, arrays 0..mp-1	*/


int ypos;			/* Global y position for graph	*/
int yspace   = 12;		/* X/Y spacing and node size	*/
int xspace   = 20;
int nodesize = 2;


char *psgrays[5] = {"black", "darkgray", "gray", "lightgray", "white" };

#define PSTRICKS 1
#define COMDRAW  2

char outformat = PSTRICKS;	/* Global output format		*/


/* Forward declarations */

nodeptr parse_tree();
void    print_tree(nodeptr n, int depth, char first,
		   char verbose, char procverbose);
int     tree_depth(nodeptr n);
int	tree_width(nodeptr n);
void    print_tree_graphics(nodeptr n, int depth,
	   		    char first, char last, char* branches,
			    char verbose);



/* Parse global input text recursively:
 * "(" opens a new node and descends into it, ")" returns from a node
 */
nodeptr parse_tree() {
  char     c;			/* Current character in input	*/
  int      count;		/* Count children		*/
  nodeptr  n;			/* This node			*/
  nodeptr *cs;			/* Child pointer for this node	*/
  int      maxchildren;		/* Currently allocated #children*/
  int      procnum = 0;		/* Procedure code-number 	*/
  char	   haveprocnum = 0;	/* True procnum	already read	*/
  long     cputime = 0;		/* ms CPU Time spent in subtree	*/


  n = (nodeptr) malloc(sizeof(node));
  n->numchildren = 0;		/* Init values			*/
  n->children = (nodeptr*)0;
  n->numleafs = 0;
  n->numnodes = 0; 
  n->procnum  = 0;
  n->cputime  = 0;
  n->info = NULL;
  n->procleafs = (int*) NULL;	/* Procedure-specific counters	*/
  n->procnodes = (int*) NULL;	/*  are filled in later		*/
  

  while (c = *(text++)) {	/* Read next character		*/

    if (c == '(') {		/* New child, descend		*/

      (n->numchildren)++;
      if (n->children == 0) {
	
	/* Didn't have children yet, allocate some space	*/

	n->children = (nodeptr*)malloc(CHILDNUM * sizeof(nodeptr));
	maxchildren = CHILDNUM;	/* Remember current max.alloc.	*/

      } else if (n->numchildren >= maxchildren) {
	
	/* More children than already allocated, make room	*/

	maxchildren = maxchildren + CHILDNUM;
	n->children = (nodeptr*)realloc( (void*) n->children,
					 maxchildren * sizeof(nodeptr));
      }

      n->children[n->numchildren-1] = parse_tree();
      n->numleafs = n->numleafs +		/* Count leafs	*/
	            n->children[n->numchildren-1]->numleafs;
      n->numnodes = n->numnodes +		/* Count nodes	*/
	            n->children[n->numchildren-1]->numnodes;

    } else if (c == ')') {	/* End of node, null		*/
      
      /* Shrink children to necessary number, if any exist	*/
      /* If not, it's a leaf, so add itself to its subtree	*/

      if (n->children != 0) {
	(n->numnodes)++;	/* This node itself is no leaf	*/
	n->children = (nodeptr*)realloc( (void*) n->children,
					 n->numchildren * sizeof(nodeptr));
      } else {
	n->numleafs = 1;
	n->numnodes = 0;
      }
      n->procnum = procnum;	/* Remember procnum/cpu time if found	*/
      n->cputime = cputime;
      if (procnum > maxprocnum)	/* Update global maximum-proc.-index	*/
	  maxprocnum = procnum;

      return n;

    } else if ((c >= '0') && (c <= '9')) {

      /* Parse procedure number resp. CPU time */

      if (haveprocnum) {	/* Procnum read, now comes the CPU time */
	cputime = cputime * 10 + (long) (c - '0');
      } else {			/* Still in procnum			*/
	procnum = procnum * 10 + (int) (c - '0');
      }


    } else if (c == ':') {

      haveprocnum = 1;		/* End of procedure number seen		*/

    } else {

      printf("parse_tree: unexpected character '%c'\n", c);

    }

  }

  if (procnum > maxprocnum)	/* Update global maximum-proc.-index	*/
      maxprocnum = procnum;
  
  n->procnum = procnum;		/* Remember procnum/cpu time if found	*/
  n->cputime = cputime;
  return n;

}


/* Collects procedure-specific information (number of leafs/nodes
 * per procedure type, leaf = no other procedures of this type in subtree)
 */
void compute_proc_infos(nodeptr n) {
    int i;
    int t;
    
    /* Allocate & initialize space for procedure-specific statistics	*/

    if (maxprocnum == 0) {
/*	printf("warning - compute_proc_infos called but no procedure numbers given\n");*/
	return;
    }
    n->procnodes = (int*) malloc(maxprocnum * sizeof(int));
    n->procleafs = (int*) malloc(maxprocnum * sizeof(int));
    for (t=0; t<maxprocnum; t++) {
	n->procnodes[t] = 0;
	n->procleafs[t] = 0;
    }

    for (i=0; i<n->numchildren; i++) {	/* Get Infos recursively	*/
	compute_proc_infos(n->children[i]);
    }
    for (t=0; t<maxprocnum; t++) {	/* Compute #s for each proctype */
	for (i=0; i<n->numchildren; i++) {
	    n->procnodes[t] = n->procnodes[t] +
		              n->children[i]->procnodes[t];
	    n->procleafs[t] = n->procleafs[t] +
		              n->children[i]->procleafs[t];
	}
	if (n->procnum-1 == t) {	/* Regard the currend proc.type	*/
	    if (n->procleafs[t] == 0) {	/* No sub-leafs? Then this one	*/
		n->procleafs[t] = 1;	/*  is a "leaf" of this proctype*/
	    } else {
		(n->procnodes[t])++;	/* Else one more node		*/
	    }
	}
    }

}
    
/* Outputs leaf/node counts for each procedure, (12 chars for each proc)
 */
void print_proc_infos(nodeptr n) {
    int t;

    /* (here, t runs from 0 to maxprocnum-2, since maxprocnum is
     * always >0 if procnums are given at all, so index 0 means 1 etc)
     */
    for (t=0; t<maxprocnum; t++) {
	if ((n->procnodes) && (n->procleafs)) {		/* (sanity check) */
	    printf("%c|%4d/%-4d ", (char) ('A'+t),
		   n->procnodes[t], n->procleafs[t]);
	} else {
	    printf("(procinfos?)");
	}
    }
}


/* Outputs the tree semi-graphically with the number of
 * leafs and nodes in the subtree. If verbose is true,
 * also output procedure number (as "A"..."Z") and CPU time used.
 * Setting procverbose to true results in subtree-node/leaf output
 * for each recursive procedure type.
 */
void print_tree(nodeptr n, int depth, char first,
		char verbose, char procverbose)
{
    int t;
    int i;
    int j;
    char any;
    char f;

    if (!first)
      for (t=0; t< depth; t++) {        /* Indent for depth     	*/
	  if ((verbose==1) && (procverbose==0)) {
	      printf("                   ");
	} else if ((verbose==0) && (procverbose==1)) {
	  printf("            ");
	  for (j=0; j<maxprocnum; j++)	/* Room for procverbose		*/
	      printf("            ");
	} else if ((verbose==1) && (procverbose==1)) {
	  for (j=0; j<maxprocnum; j++)	/* Room for procverbose		*/
	      printf("            ");
	} else {
	  printf("            ");
	}
      }
    
    if (verbose && !procverbose) {
	printf(" %c:%4dms %4d/%-4d", (char) ('A'-1+n->procnum),
	       n->cputime, n->numnodes, n->numleafs);
    } else if (!verbose && procverbose) {
	printf(" %5d/%-5d", n->numnodes, n->numleafs);
	print_proc_infos(n);
    } else if (verbose && procverbose) {
	printf(" %c:%4dms %4d/%-4d", (char) ('A'-1+n->procnum),
	       n->cputime, n->numnodes, n->numleafs);
	print_proc_infos(n);
    } else { 
	printf(" %5d/%-5d", n->numnodes, n->numleafs);
    }
    f=1;

    for (i=0; i<n->numchildren; i++) {
      print_tree(n->children[i], depth+1, f, verbose, procverbose);
      f=0; 	                   	/* Reset flag after 1st print	*/
    }
    if (!n->numchildren)
      printf("\n");             	/* New line if no children  	*/
    
}


/* Print tree graphically using ASCII output.
 * If verbose flag is set, print procedure number (A,B,...)
 * instead of node symbol.
 */
void print_tree_graphics(nodeptr n, int depth,
			 char first, char last, char* branches,
			 char verbose)
{
    int t;
    int i;
    char f;	/* Flag: True if first child of this node	*/
    char l;	/* Flag: True if last child of this node	*/
    char any;

    if (!first) {
      for (t=0; t < depth; t++)		/* Indent for depth	*/
	  printf("%c  ", branches[t]);
    }
    
    f=1;


    for (i=0; i<n->numchildren; i++) {	/* Lengthen branch if any subbranches*/
	l = (char)(i >= n->numchildren-1);
	if (f) {
	    if (verbose)		/* Procedure number as character*/
	        printf("%c--", (char)(n->procnum + 'A'-1));
	    else
	        if ((last) && (!first))
		    printf("`--");	/* Last branch looks different	*/
	        else	
		    printf("+--");
	    branches[depth+1]='|';
	}				/* Set flag for last child	*/
	if (l)				/* Only reset after last child	*/
	    branches[depth+1]=' ';
	print_tree_graphics(n->children[i], depth+1,
			    f, l, branches, verbose);
	f=0;
    }
    if (!n->numchildren)		/* Print leaf correctly		*/
        if (verbose) {
	    printf("%c\n", (char)(n->procnum + 'A'-1));
	} else {
	    if ((last) && (!first))
	        printf("`-O\n"); 
	    else
	        printf("+-O\n"); 
	}

}

int maxx = 0;
int maxy = 0;


/* Sets global maxx/maxy to the maximum dimensions reached when
 * drawing the tree on a 2D surface (as the drawcommand stuff below
 * does it). Used for setting the pspicture drawing size.
 * Affects the global variable ypos!
 */
void tree_size(nodeptr n, int depth)
{
    int  i;
    char f;     /* Flag: True if first child of this node       */

    f=1;
    for (i=0; i<n->numchildren; i++) {
	if (!f) {
	    ypos++;			/* Don't increase ypos for 1st child */
	}
	tree_size(n->children[i], depth+1);
	f=0;
    }

    if (ypos*yspace + nodesize > maxy) maxy=ypos*yspace + nodesize;
    if ((depth+1)*xspace + nodesize > maxx) maxx=(depth+1)*xspace + nodesize;
}



void drawcommand_init(nodeptr n)
{
    if (outformat == COMDRAW) {
	printf("drawtool(\n");

	/* Initialize graphics with dummy point */
    
	printf("zoomout(),\n");
	printf("dummy=line(%d,%d,%d,%d),\n", xspace, yspace, xspace, yspace);
	printf("select(dummy),\n");
	printf("colors(1,10),\n");
	printf("pattern(2),\n");
	printf("brush(3),\n");		/* 1 pixel thick	*/
    } else {
	
	printf("\\psset{unit=0.5mm}\n");
	printf("\\psset{linewidth=0.1}\n");
	printf("\\psset{linecolor=black}\n");
	printf("\\psset{fillstyle=solid}\n");
	printf("\\psset{dimen=middle}\n");
	printf("\n");

	ypos = 1;
	maxx=maxy=0;
	tree_size(n, 1);
	
	printf("\\begin{pspicture}(0,0)(%d,%d)\n",
	       maxx, maxy);
	
    }
}


void drawcommand_end()
{
    if (outformat == COMDRAW) {
	printf("quit()\n");
	printf(")\n");
    } else {
	printf("\\end{pspicture}\n");
    }
}


void drawcommand_line(int x1, int y1, int x2, int y2)
{
    if (outformat == COMDRAW) {
	printf("dummy=line(%d,%d,%d,%d),\n",	/* Draw line		     */
	       x1, y1, x2, y2);
	if (nodesize>0) {
	    printf("select(dummy),\n");		/* Select freshly drawn line */
	    printf("pattern(2),\n");		/* Ensure pattern (only	     */
	}					/*  needed if nodes drawn    */
    } else {
	printf("\\qline(%d,%d)(%d,%d)\n",
	       x1, y1, x2, y2);
    }
}


void drawcommand_node(int x, int y, int size, int color)
{
    if (outformat == COMDRAW) {
	printf("dummy=ellipse(%d,%d,%d,%d),\n",	/* Draw node		     */
	       x, y, size, size);
	printf("select(dummy),\n");		/* Select freshly drawn node */
	printf("pattern(%d),\n",		/* (gray % = procnum)	     */
	       color+2);
    } else {
	printf("\\pscircle[fillcolor=%s](%d,%d){%d}\n",
	       psgrays[color],
	       x, y, size);
    }
}


void print_drawcommand_subtree(nodeptr n, int depth)
{
    int  i;
    char f;     /* Flag: True if first child of this node       */
    int  myypos = ypos;

    
    f=1;

    for (i=0; i<n->numchildren; i++) {
	if (!f) {
	    ypos++;			/* Don't increase ypos for 1st child */
	}
	drawcommand_line(		/* Line to child	     */
	    depth*xspace, myypos*yspace,
	    (depth+1)*xspace, ypos*yspace);
	
	print_drawcommand_subtree(n->children[i], depth+1);
	f=0;
    }

    /* Draw node last to make it topmost object.
     * Only draw node if size > 0 (trees with only branches are OK, too)
     */

    if (nodesize>0) {
	drawcommand_node(		/* Draw node		     */
	    depth*xspace, myypos*yspace,	
	    nodesize, (n->procnum)%5);
    }

}


/* Output tree in graphical command language format (comdraw/pstricks)
 */
void print_drawcommand_tree(nodeptr n)
{
    ypos = 1;

    drawcommand_init(n);
    ypos = 1;
    print_drawcommand_subtree(n, 1);
    drawcommand_end();
    
}



/* Compute depth of tree recursively
 * (a one-leaf-only tree has depth 1 but its node(leaf) will be at
 * level 0, i.e. the depths of the tree range from 0..tree_depth(root) !)
 */
int tree_depth(nodeptr n)
{
    int i;
    int d;
    int cd;

    d=1;				/* Own depth-increase		*/
    for (i=0; i<n->numchildren; i++) {
	cd = tree_depth(n->children[i]);
	if (cd+1 > d)			/* Remember maximum child depth	*/
	    d=cd+1;
    }
    return d;
   
}


/* Compute maximum branching degree of tree recursively
 */
int tree_degree(nodeptr n)
{
    int i;
    int d;
    int cd;

    d=n->numchildren;			/* Own degree			*/
    for (i=0; i<n->numchildren; i++) {
	cd = tree_degree(n->children[i]);
	if (cd > d)			/* Remember maximum child degree*/
	    d=cd;
    }
    return d;
   
}


main(int argc, char **argv)
{
  char    tree_out       = 0;	/* Output tree?				*/
  char	  tree_verbose   = 0;	/* Verbose tree output (timings etc) ?	*/
  char	  grtree_out     = 0;	/* Ditto graphical tree			*/
  char	  grtree_verbose = 0;
  char    stats_out      = 0;	/* Ditto statistics			*/
  char	  stats_verbose  = 0;
  char    debug_out      = 0;	/* More detailed output			*/
  
  nodeptr root;
  nodeptr *subroot;
  char   *inputtext = (char*) malloc(1);
  char    inputbuf[TEXTCHUNKSIZE];
  int     inputlen = 0;
  int     l;
  int     i,j;
  int     depth;
  nodeptr n;
  char    branchtext[BUFSIZ];

  for (i=1; i<argc; i++) {
      if (!strcmp(argv[i],"-x")) {
          if (i==argc-1) {
              printf("Error, -x needs argument (x spacing)\n");
              exit(10);
          }
          i++;
	  xspace = atoi(argv[i]);
	  if (xspace<1) {
              printf("Error, x spacing must be > 0\n");
              exit(11);
          }
      } else if (!strcmp(argv[i],"-y")) {
          if (i==argc-1) {
              printf("Error, -y needs argument (y spacing)\n");
              exit(20);
          }
          i++;
	  yspace = atoi(argv[i]);
	  if (yspace<1) {
              printf("Error, y spacing must be > 0\n");
              exit(21);
          }
      } else if (!strcmp(argv[i],"-s")) {
          if (i==argc-1) {
              printf("Error, -s needs argument (node size)\n");
              exit(30);
          }
          i++;
	  nodesize = atoi(argv[i]);
	  if (nodesize<0) {
              printf("Error, node size must be >= 0\n");
              exit(31);
          }
      } else if ((!strcmp(argv[i],"-pstricks")) ||
		 (!strcmp(argv[i],"-p")) ) {
	  outformat = PSTRICKS;
      } else if ((!strcmp(argv[i],"-comdraw")) ||
		 (!strcmp(argv[i],"-c")) ) {
	  outformat = COMDRAW;
      } else {
	  printf("%s - reads profile-generated tree from stdin,\n", argv[0]);
	  printf("   writes tree graphics as commands to stdout.\n");
 	  printf("   Usage: tree_graph_output [options] < treefile > comfile\n");
	  printf("   Options:\n");
	  printf("     -comdraw  Output in comdraw format  (also '-c')\n");
	  printf("     -pstricks Output in pstricks format (also '-p', default)\n");
	  printf("     -x n      Sets x spacing of nodes to n (default=20)\n");
	  printf("     -y n      Sets y spacing of nodes to n (default=12)\n");
	  printf("     -s n      Sets node diameter to n      (default= 2)\n");
	  printf("     treefile  ASCII representation of recursion tree, e.g.\n");
	  printf("               \"(()())\" for simple binary tree, may be annotated\n");
	  printf("               with procedure numbers and subtree timings \n");
	  exit(1);

      }
  }
  
  *inputtext = '\0';
  while (feof(stdin) == 0) {
      if (fgets(inputbuf, TEXTCHUNKSIZE, stdin)) {
	  l = strlen(inputbuf);
	  inputlen = inputlen + l;
	  inputtext = (char*)realloc(inputtext, inputlen);
/*	  printf("read %d bytes, chunk size = %d\n", l, TEXTCHUNKSIZE);*/
	  strcat(inputtext, inputbuf);
      }
  }
  if (strlen(inputtext)>1)
      text = inputtext+1;	/* Skip first dummy-root-node		*/
  else
      text = inputtext;

  root = parse_tree();

  compute_proc_infos(root);	/* Get subtree nodes/leafs for each proc*/
  
  print_drawcommand_tree(root);	/* Output as graphical commands		*/
  
}

